use std::{
    fs::{self, OpenOptions},
    io::{ErrorKind, Write},
};

use anyhow::Result;

#[derive(Debug)]
pub struct Store {
    path: String,
}

impl Store {
    pub fn open(path: impl ToString) -> Result<Self> {
        let path = path.to_string();
        fs::create_dir_all(&path)?;
        Ok(Self { path })
    }

    pub fn list(&self) -> Result<Vec<String>> {
        let mut results = Vec::new();
        for result in fs::read_dir(&self.path)? {
            let entry = result?;
            let name = entry.file_name().into_string().unwrap();
            results.push(name);
        }
        results.sort();
        Ok(results)
    }

    pub fn activity(&self) -> Result<Vec<String>> {
        let mut results = Vec::new();
        for result in fs::read_dir(&self.path)? {
            let entry = result?;
            let name = entry.file_name().into_string().unwrap();
            let modif_time = entry.metadata()?.modified()?;
            let duration = modif_time.duration_since(std::time::UNIX_EPOCH)?;
            let date = duration.as_nanos();
            results.push((name, date));
        }

        results.sort_by_key(|&(_, date)| date);
        Ok(results.into_iter().rev().map(|(name, _)| name).collect())
    }

    pub fn get(&self, topic: &str) -> Result<Vec<String>> {
        let base = &self.path;
        let path = format!("{base}/{topic}");
        let content = fs::read_to_string(path)?;
        Ok(content.split(SEPARATOR).map(String::from).collect())
    }

    pub fn insert(&mut self, topic: &str, content: String) -> Result<()> {
        let content = content.replace('\u{c}', " ");
        let content = content.trim();
        let base = &self.path;
        let path = format!("{base}/{topic}");
        match OpenOptions::new().append(true).create(false).open(&path) {
            Ok(mut file) => write!(&mut file, "{SEPARATOR}\n{content}\n")?,
            Err(error) if error.kind() == ErrorKind::NotFound => drop(fs::write(path, content)),
            Err(error) => Err(error)?,
        };
        Ok(())
    }
}

const SEPARATOR: &str = "\n\u{c}\n";
